#!/usr/bin/env python
from __future__ import print_function
"""
Copyright 2015 Hippos Technical Systems BV.
(but borrows somme code from the painting/svgviewer example of PyQt v4.x)
DECLARED OBSOLESCENT per 4th April 2016; see PitchView.py
@author: larry
"""

import sys, re
import numpy as np
from PIL import Image
from ..share import (Share, dbg_print, QtCore, QtGui, image_dir, QtSvg, WithMenu, Printer)
import pyqtgraph as pg
from .terpsichore import default_voice, Clef, Note


# class MyScene(pg.GraphicsScene):
# etc... removed; not working; not necesary?


class FreqView(pg.GraphicsView, WithMenu):
    menuTag = '&Tuning'
    pps = 1000  # points per semitone
    n_samples = 1000

    def menuItems(self):
        return [
        ]

    def __init__(self):
        dbg_print ("FreqView.__init__")
        pg.GraphicsWindow.__init__(self)
        WithMenu.__init__(self)
        # self.setScene(MyScene(self)) # see above

        # avoid showing "window in the making"' during startup.
        # not sure when or why window gets shown again but it does so conveniently!\
        self.hide()
        self.plotStaves()
        self.data = np.array([np.NaN]*self.n_samples, dtype=np.float32)
        self.curve = self.staves_plot.plot(self.data, axisItems={}) # , connect='finite')
        dbg_print ("!FreqView.__init__")

    def plotStaves(self):
        self.staves_plot = self.addPlot()
        self.staves_plot.hideAxis('left')
        self.staves_plot.setYRange(36*self.pps, 80*self.pps)
        #self.staves_plot.setXRange(-self.n_samples, 0)
        self.staves_plot.setXRange(0, self.n_samples)

        all_staff_lines = []  # accumulator
        for clef in (Clef.Bass, Clef.Treble):
            all_staff_lines += clef.lines
            img = Image.open(image_dir + clef.symbol).resize(
                (32*clef.scaleHint, 920*clef.scaleHint), Image.ANTIALIAS)
            img_ar = np.array(img)
            #print(img_ar.shape)
            img_ar_v = np.swapaxes(np.flipud(img_ar.view(dtype=np.uint32).reshape(img_ar.shape[:-1])), 0, 1)
            img_item = pg.ImageItem(img_ar_v, opacity=0.5)
            self.staves_plot.addItem(img_item, autoLevels=False)

            marked_semi = clef.lines[clef.marked]
            img_item.setPos(0, (marked_semi.GetPitch() - clef.descent)*self.pps)

        for ixPitch in range(128):
            note = default_voice.GetNote(ixPitch)
            width = 1
            style = None
            label = note.real_name
            labelOpts={'color': (200,0,0), 'movable': False, 'position': 0.8,
                       'anchors': (0.5, 0.5), 'fill': (0, 0, 200, 100)}
            if note in all_staff_lines:
                colour = (200, 200, 200)
                width = 2
            #elif note is Note.A4: # traditional tuning note nominally 440Hz but often inflated
            #    colour = (0, 200, 0)
            #elif note.pitch == 0:
            elif note is Note.C4:
                label += " = middle C"
                style = QtCore.Qt.DashLine
                colour = (220, 220, 220)
            elif len(note.real_name) == 2:
                colour = (200, 30, 30)
            else:
                colour = (150, 30, 30)
            inf_line = pg.InfiniteLine(movable=False, angle=0, pos=ixPitch*self.pps,
                                       pen=dict(color=colour, width=width, style=style),
                                       # bounds = [-20, 20],
                                       hoverPen=(0,240,0),
                                       label=label, labelOpts=labelOpts)
            self.staves_plot.addItem(inf_line)

    def add_sample(self, volume, freq):
        self.data[:-1] = self.data[1:]
        self.data[-1] = freq is None and np.NaN or (freq * self.pps)
        self.curve.setData(self.data)  # , connect='finite')

    def wheelEvent(self, event):
        factor = 1.2**( event.delta() / 120.0)
        self.scale(factor, factor)
        # self.mustApplyTransform = self.transform()
        dbg_print ("Score.wheelEvent, delta = ", event.delta())
        event.accept()



#!/usr/bin/env python
"""                                                                            
Play MIDI file on output port.

Run with (for example):

    ./play_midi_file.py 'SH-201 MIDI 1' 'test.mid'
"""

import sys, mido
from ..share import (Share, Signal, WithMenu, dbg_print, QtCore, QtGui)


class MidiPlayer(QtGui.QWidget):
        
    lineAndCol = Signal(int, int)

    outputPort = 'TiMidity port 0'
    output = None

    def __init__(self):
        QtGui.QWidget.__init__(self)
        # self.open_output()

    def open_output(self):
        try:
            self.output = mido.open_output(self.outputPort)
        except IOError as exc:
            print ("sorry; couldn't setup MIDI player; exception details follow <<<")
            print (exc)
            print (">>>. Perhaps overruling outputPort (currently {0}) will help".format(self.outputPort))
            self.output = None
        dbg_print('MidiPlayer:__init__', self.output)

    def play(self, filename):
        if self.output:
            self.output.reset()
        else:
            self.open_output()
        self.accum = dict([(i, 0) for i in range(110, 115)])
        self.midiFile = mido.MidiFile(filename)
        self.messages = self.midiFile.__iter__() # self.midiFile.play()
        self.pendingMessage = None
        self.paused = False
        self.cueMessage()

    def pause(self):
        self.paused = not self.paused
        self.cueMessage()

    def cueMessage(self):
        if self.paused:
            return
        message = self.pendingMessage
        self.pendingMessage = None
        while True:
            if message:
                if not isinstance(message, mido.MetaMessage):
                    if message.type == 'control_change':
                        self.accum[message.control] = message.value
                    else:
                        self.output.send(message)
                        lineNo = ((self.accum[110]<<14)
                                 +(self.accum[111]<<7)
                                 + self.accum[112])
                        colNo =  ((self.accum[113]<<7)
                                 + self.accum[114])
                        self.lineAndCol.emit(lineNo, colNo-1)
            try:
                message = next(self.messages)
            except StopIteration:
                dbg_print('cue_msg; StopIteration')
                return
            dbg_print(message.type, message)
            if message.time != 0:
                break
        self.pendingMessage = message
        milliseconds = int(message.time * 1000)
        QtCore.QTimer.singleShot(milliseconds, self.cueMessage)

    def __del__(self):
        dbg_print('MidiPlayer:__del__',)
        #self.output.reset()

if __name__ == '__main__':
    class MainWindow(QtGui.QMainWindow):
        def __init__(self):
            super(MainWindow, self).__init__()
    
            self.midiPlayer = MidiPlayer()
    
            fileMenu = QtGui.QMenu("&File", self)
            quitAction = fileMenu.addAction("E&xit")
            quitAction.setShortcut("Ctrl+Q")
    
            self.menuBar().addMenu(fileMenu)
    
    
            quitAction.triggered.connect(QtGui.qApp.quit)
    
            self.setCentralWidget(self.midiPlayer)
            self.setWindowTitle("MidiPlayer")
            self.midiPlayer.lineAndCol.connect(self.showLocation)
            self.midiPlayer.play(
                (len(sys.argv)>1 and sys.argv[1]) or './MarThruAm_timp.midi')
            #self.resize(self.view.sizeHint() + QtCore.QSize(
            #    80, 80 + self.menuBar().height()))

        def showLocation(self, lineNo, colNo):
            dbg_print('showLocation(line, col)', lineNo, colNo)

    app = QtGui.QApplication(sys.argv)
    mainWindow = MainWindow()
    mainWindow.show()
    sys.exit(app.exec_())



#!/usr/bin/env python
# -*- encoding: utf8 -*-
from __future__ import print_function
"""
Copyright 2015 Hippos Technical Systems BV.
(but borrows somme code from the painting/svgviewer example of PyQt v4.x)

@author: larry
"""

import sys, os, re
import lxml.etree
import numpy as np

from ..share import (Share, dbg_print, QtCore, QtGui, QtWebKit, WithMenu, Printer)


class HtmlView(QtWebKit.QWebView, WithMenu):

    menuTag = '&Html'

    def menuItems(self):
        return [
                    ('&Print',         None,  self.printAll,),
                    ('E&xport to PDF', None,  self.printAllToPDF,),
        ]

    def __init__(self):
        dbg_print ("HtmlView.__init__")
        QtWebKit.QWebView.__init__(self)
        WithMenu.__init__(self)

    def showOutput(self, html_bytes):
        self.setHtml(html_bytes)
        with open(self.fileName, 'w') as output:
            output.write(html_bytes)

    def showAtRowAndCol(self, row, colhtml_bytes):
        pass  # for now!

    def locateXY(self, x, y):
        pass  # for now!

    def printAll(self, toPDF=False):
        dbg_print("printAll!")
        self.printer.setPageSize(QtGui.QPrinter.A4)
        self.printer.setOutputFormat(QtGui.QPrinter.PdfFormat)
        #self.printer.setOutputFileName(toPDF and self.compositeName or '')
        # see quick fix in .external! ...
        printName = toPDF and (os.path.splitext(self.fileName)[0] + '.pdf') or ''
        printName = printName.replace('autosave_', '')
        dbg_print("printing to file ", printName)
        self.printer.setOutputFileName(printName)
        self.print_(self.printer)



#!/usr/bin/python
# -*- encoding: utf8 -*-
"""
Copyright 2015 Hippos Technical Systems BV.

"""
from __future__ import print_function
import os, re,time
from ..share import (Share, dbg_print, QtCore, QtGui)


class SliderPlus(QtGui.QGroupBox):
    _minimum = 25
    _maximum = 400
    _value = 100
    single_step = 1
    tick_interval = 10
    tick_position = QtGui.QSlider.TicksBothSides
    _orientation = QtCore.Qt.Vertical
    focus_policy = QtCore.Qt.StrongFocus

    def __init__(self, orientation, title, parent=None, client=None):
        QtGui.QGroupBox.__init__(self, title, parent)
        self._client = client
        self.slider = QtGui.QSlider(orientation)
        self.slider.setFocusPolicy(self.focus_policy)
        self.slider.setTickPosition(self.tick_position)
        self.slider.setTickInterval(self.tick_interval)
        self.slider.setSingleStep(self.single_step)
        self.slider.setMinimum(self._minimum)
        self.slider.setMaximum(self._maximum)
        self.slider.setValue(100)
        self.slider.valueChanged.connect(self.reactToSlide)

        if orientation == QtCore.Qt.Horizontal:
            direction = QtGui.QBoxLayout.TopToBottom
        else:
            direction = QtGui.QBoxLayout.LeftToRight

        slidersLayout = QtGui.QBoxLayout(direction)
        slidersLayout.addWidget(self.slider)
        self.setLayout(slidersLayout)

    def write_to_client(self, s):
        if not self._client:
            return
        self._client.feed_input(s)

    def reactToSlide(self, value):
        print('slid to', value)
        self.adjust(value)

    def adjust(self, value):
        print("don't call me directly")

    def setValue(self, value):
        if not self.slider.isSliderDown():
            self.slider.setSliderPosition(value)


class SpeedChanger(SliderPlus):
    def adjust(self, value):
        #if self._client:
        #    # self._client.speed = value / 100.0
        speed =  value / 100.0
        self.write_to_client("speed_set %.2f" % speed)

class PosSeeker(SliderPlus):
    _minimum = 0
    _maximum = 100
    _value = 100
    tick_interval = 10

    def adjust(self, value):
        if self._client:
            self._client.percent_pos = value


class PlayerControl(QtGui.QWidget):
    headerText = 'player'
    whereDockable = QtCore.Qt.AllDockWidgetAreas
    menu = None

    def __init__(self, parent=None, dock=None, client=None):
        QtGui.QWidget.__init__(self, parent=parent)
        self._client = client
        self.speed_changer = SpeedChanger(QtCore.Qt.Horizontal, "Speed", client=client)
        self.pos_seeker = PosSeeker(QtCore.Qt.Horizontal, "Position", client=client)
        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.speed_changer)
        layout.addWidget(self.pos_seeker)
        self.setLayout(layout)
        # self.setWindowTitle("Sliding")
        # parent.timer.timeout.connect(self.trackPos)

    def trackPos(self):
        percentPos = self._client.percent_pos
        print("trackPos", percentPos)
        # self.pos_seeker.setValue(percentPos)



# -*- coding: utf-8 -*-
"""
Various methods of drawing scrolling plots.
"""
import sys, os
import pyqtgraph as pg
from ..share import (dbg_print, QtCore, QtGui, Printer, image_dir)
import numpy as np
import math
import pyaudio
from PIL import Image
from .terpsichore import default_voice, Clef, Note
Signal = QtCore.Signal

SHOW_VOLUMES = 0
AUDIO_SAMPLE_RATE = 44100


class PitchPlot(pg.PlotItem):

    burst_signal = Signal(int, np.ndarray)

    fft_pen = (220, 220, 42) # bright yellow
    note_trace_pen = (128, 240, 128) # bright green?
    concertPitch=440.0
    pps = 1000  # points per semitone
    trace_scale = 0.005  # just a magic number 'for now'
    MIN_SAMPLES = AUDIO_SAMPLE_RATE * 0.02  # don't attempt to derive freq on less than 20 milliseconds

    def __init__(self):
        pg.PlotItem.__init__(self)
        self.setLabel('top', 'amplitude', '') # , pen=self.fft_pen)
        self.setLabel('bottom', 'Time', 's')
        self.setLabel('left', 'midi note', 'mst')

        self.setYRange(30*self.pps, 96*self.pps)
        self.setXRange(0, 2000)
        self.fft_curve = self.plot() # np.arange(1280), np.arange(1280))  # self.recent_samples)
        self.fft_curve.setPen(self.fft_pen)

        #self.recent_notes = np.array([60000]+[np.NaN]*10000)
        self.recent_notes = np.arange(2000) * 30.
        self.note_trace = self.plot(self.recent_notes, connect='finite')
        self.note_trace.setPen(self.note_trace_pen)
        self.draw_staves_etc()
        self.setPreferredSize(800, 600)
        self.it = 0
        self.burst_signal.connect(self.interpret_burst)

    def draw_staves_etc(self):
        all_staff_lines = []  # accumulator

        for clef in (Clef.Bass, Clef.Treble):
            all_staff_lines += clef.lines
            img = Image.open(image_dir + clef.symbol).resize(
                (32*clef.scaleHint, 900*clef.scaleHint), Image.ANTIALIAS)
            img_ar = np.array(img)
            img_ar_v = np.swapaxes(np.flipud(img_ar.view(dtype=np.uint32).reshape(img_ar.shape[:-1])), 0, 1)
            img_item = pg.ImageItem(img_ar_v, opacity=0.5)
            self.addItem(img_item, autoLevels=False)

            marked_semi = clef.lines[clef.marked]
            img_item.setPos(0, (marked_semi.GetPitch() - clef.descent)*self.pps)

        for ixPitch in range(128):
            note = default_voice.GetNote(ixPitch)
            width = 1
            style = None
            label = note.real_name
            labelOpts={'color': (200,0,0), 'movable': False, 'position': 0.8,
                       'anchors': (0.5, 0.5), 'fill': (0, 0, 200, 100)}
            if note in all_staff_lines:
                colour = (200, 200, 200)
                width = 2
            #elif note is Note.A4: # traditional tuning note nominally 440Hz but often inflated
            #    colour = (0, 200, 0)
            #elif note.pitch == 0:
            elif note is Note.C4:
                label += " = middle C"
                style = QtCore.Qt.DashLine
                colour = (220, 220, 220)
            elif len(note.real_name) == 2:
                colour = (200, 30, 30)
            else:
                colour = (150, 30, 30)
                label = None
            inf_line = pg.InfiniteLine(movable=False, angle=0, pos=ixPitch*self.pps,
                                       pen=dict(color=colour, width=width, style=style),
                                       # bounds = [-20, 20],
                                       # hoverPen=(0,240,0),
                                       ) #label=label, labelOpts=labelOpts)
            self.addItem(inf_line)

    def interpret_burst(self, quiets, samples, believe=True, minFreq=10, maxFreq=1024, spread=3,
                            transpose=False):
        nSamples = len(samples)
        # print ('interpret_burst', quiets, nSamples)
        if len(samples) < self.MIN_SAMPLES:
            quiets += len(samples)
            samples = samples[0:0]  # empty it but keep the type
            # print ("tiddler!")
        if len(samples):
            self.fftAmplitudes = np.array((abs(np.fft.fft(samples))/nSamples)[minFreq:maxFreq],
                                          dtype=np.int32)
            self.fftFrequencies = (AUDIO_SAMPLE_RATE*np.arange(nSamples))[minFreq:maxFreq]/(nSamples*1)
            indexMaxAmplitude = np.argmax(self.fftAmplitudes)
            amplitudes = self.fftAmplitudes[indexMaxAmplitude-spread:indexMaxAmplitude+spread+1]
            frequencies = self.fftFrequencies[indexMaxAmplitude-spread:indexMaxAmplitude+spread+1]
            total = np.sum(amplitudes)
            volume = math.log(total+1, 2.0)
            avgFrequency = np.sum(frequencies*amplitudes)/total
            absPitch = 1000 * (69 + 12 * math.log((avgFrequency / self.concertPitch), 2.0))
            # print("absPitch =", absPitch, 'mst')
            try:
                avgNote = default_voice.GetNote(absPitch / 1000., transpose=False)
            except (ValueError, IndexError):
                avgNote = np.NaN
                absPitch = None
            # print(avgNote)

            pitches = self.pps * (np.log2(self.fftFrequencies / self.concertPitch)*12+69)
            self.fft_curve.setData(self.fftAmplitudes * 10, pitches)
        else:
            avgNote = np.NaN
            absPitch = None
        l = int((quiets+len(samples)) * self.trace_scale)
        # absPitch = self.it * 1000.
        self.recent_notes[:-l] = self.recent_notes[l:]
        self.recent_notes[-l:] = np.array([absPitch or np.NaN]*l)
        self.note_trace.setData(self.recent_notes, connect='finite')
        self.update()
        self.show()


class PitchView(pg.GraphicsView):
    FRAMES_PER_BUFFER = 1024
    FORMAT = pyaudio.paInt16
    CHANNELS = 1
    INPUT_DEVICE_INDEX = 2 # None  # i.e use default
    MIN_SIG_VOLUME = 400
    MAX_BURST = AUDIO_SAMPLE_RATE * 0.2
    MAX_QUIETS = AUDIO_SAMPLE_RATE * 0.5

    update_signal = Signal(np.ndarray)

    one_shot_mode = False

    def __init__(self):
        pg.GraphicsView.__init__(self)
        self.setWindowTitle("Musicraft: Pitch and Note Plot")
        self.lay = pg.GraphicsLayout(border=(100, 100, 100))
        self.setCentralItem(self.lay)
        #raw_plot = self.lay.addPlot()
        raw_plot = pg.PlotItem()
        self.lay.addItem(raw_plot)
        raw_plot.setYRange(-1024, 1024)
        raw_plot.setMaximumSize(200., 200.)
        self.recent_samples = np.zeros(AUDIO_SAMPLE_RATE, dtype=np.int16)  # display one seconds worth of raw data
        self.raw_curve = raw_plot.plot(self.recent_samples)
        self.start_index = self.end_index = self.quiets = 0
        self.thresholds = [pg.InfiniteLine(pos=y, angle=0, pen=None, movable=True)
                           for y in (self.MIN_SIG_VOLUME, - self.MIN_SIG_VOLUME)]
        for is_other, th in enumerate(self.thresholds):
            raw_plot.addItem(th)
            th.sigPositionChanged.connect(self.threshold_changed)

        self.pitch_plot = PitchPlot()
        self.lay.addItem(self.pitch_plot, col=1, colspan=6)
        self.pitch_plot.setPreferredHeight(500.)

        self.init_audio()
        self.show()

    def threshold_changed(self, line):
        y = line.pos().y()
        for th in self.thresholds:
            if line is not th:
                th.setPos(-y)
        self.MIN_SIG_VOLUME = abs(y)


    def init_audio(self):
        self.update_signal.connect(self.accumulate)
        self.pyaud = pyaudio.PyAudio()
        self.stream = self.pyaud.open(
            format=self.FORMAT,
            channels=self.CHANNELS,
            rate=AUDIO_SAMPLE_RATE,
            frames_per_buffer=self.FRAMES_PER_BUFFER,
            input_device_index=self.INPUT_DEVICE_INDEX,
            input=True,
            stream_callback = self.got_audio_block
        )
        return

    def got_audio_block(self, raw_bytes, frame_count, time_info, status):
        dbg_print(frame_count, time_info, status)
        self.update_signal.emit(np.fromstring(raw_bytes, dtype=np.int16))
        return raw_bytes, pyaudio.paContinue

    def accumulate(self, new_samples):
        dbg_print("accumulate", len(new_samples))
        if self.end_index:  # 'update_graphs' clears this except in one-shot mode!
            return

        self.recent_samples[:-self.FRAMES_PER_BUFFER] = self.recent_samples[self.FRAMES_PER_BUFFER:]
        self.recent_samples[-self.FRAMES_PER_BUFFER:] = new_samples
        abs_samples = abs(new_samples)
        i_max_volume = np.argmax(abs_samples)
        new_volume = abs_samples[i_max_volume]
        start_of_frame_index = len(self.recent_samples) - self.FRAMES_PER_BUFFER
        on = new_volume > self.MIN_SIG_VOLUME
        if self.start_index:  # in middle of burst (or blip)
            self.start_index -= len(new_samples)
            if not on:  # end of burst
                self.end_index = start_of_frame_index
                new_quiets = self.FRAMES_PER_BUFFER
            elif (start_of_frame_index - self.start_index) > self.MAX_BURST:
                self.end_index = len(self.recent_samples)
                new_quiets = 0
        else:  # was silent(ish) at last sample
            if on:
                self.start_index = start_of_frame_index
            else:
                self.quiets += self.FRAMES_PER_BUFFER
                if self.quiets >self.MAX_QUIETS:
                    self.start_index = self.end_index = len(self.recent_samples)
                    new_quiets = 0

        if self.end_index:  # i.e if any cause to alert the burst handler...
            self.pitch_plot.burst_signal.emit(self.quiets,
                                              self.recent_samples[self.start_index:self.end_index])
            self.quiets = new_quiets
            self.start_index = 0
            if not self.one_shot_mode:
                self.end_index = 0
        self.update_amplitude_graph()

    def update_amplitude_graph(self):
        self.raw_curve.setData(self.recent_samples)

if __name__ == '__main__':
    pg.mkQApp()
    apw = PitchView()
    apw.start()
## Start Qt event loop unless running in interactive mode or using pyside.
    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()



#!/usr/bin/python
#partially based on: http://john.nachtimwald.com/2009/08/15/qtextedit-with-line-numbers/ (MIT license)
from __future__ import print_function
import sys, os, subprocess
from ..share import (Share, Signal, dbg_print, QtCore, QtGui, QtSvg, temp_dir)

##LMY: from highlighter import PythonHighlighter

class Editor(QtGui.QPlainTextEdit):

    headerText = 'Edit'
    prevCursorPos = -1
    currentLineColor = None
    editBecomesActive = Signal()
    specialSaveFileName = None
    fileName = None
    highlighter = None
    pointSizeF = 11.0
    cursorWidth = 8

    def __init__(self, book=None, **kw):
        self.book = book
        QtGui.QPlainTextEdit.__init__(self, **kw)
        self.lineNumberArea = self.LineNumberArea(self)
        self.viewport().installEventFilter(self)
        
        self.newDocument = True
        self.path = ''
        css = '''
        QPlainTextEdit {
          font-family: monospace;
          font-size: 10;
          color: black;
          background-color: white;
          selection-color: white;
          selection-background-color: #437DCD;
        }'''
        self.setStyleSheet(css)

        font = self.font()
        font.setPointSize(self.pointSizeF)
        self.setFont(font)
        self.setCursorWidth(self.cursorWidth)
        self.setWindowTitle('title')
        self.textChanged.connect(self.handleTextChanged)
        self.editBecomesActive.connect(self.handleTextChanged)
        self.setLineWrapMode(QtGui.QPlainTextEdit.NoWrap)
        self.cursorPositionChanged.connect(self.handleCursorMove)
        self.originalText = None
        self.haveLoadedFile = False

    def Quote(self):
        tC = self.textCursor()
        c0 = '#' # dummy non-match!
        while c0 not in "ABCDEFG":
            tC.movePosition(tC.Left, tC.KeepAnchor)
            sel = tC.selectedText()
            c0 = sel[0]
        tC.removeSelectedText()
        tC.insertText('"'+ sel +'"')

    def handleCursorMove(self):
        self.book.counted = self.book.latency
        return

    def moveToRowCol(self, row=1, col=0):
        block = self.document().findBlockByLineNumber (row-1)
        desiredPosition = block.position() + col
        dbg_print ('AbcEditor.moveToRowCol', row, col,
               'desiredPosition', desiredPosition)
        tc = self.textCursor()
        tc.setPosition(desiredPosition)
        self.setTextCursor(tc)
        self.setFocus()
        if self.highlighter:
            self.highlighter.rehighlight()

    def highlight(self, tc):
        # n.b. unfortunate name - no relation to highlighter!
        blockNumber = tc.blockNumber()
        # Common.blockNumber = blockNumber
        col0 =  col = tc.positionInBlock()
        l = tc.block().length()
        dbg_print ("autoTrack", l)
        blockText = tc.block().text()
        if 0:  # under review sine new approach to syntqx highlighting:
            while col and ((col >= (l-1))
                or not (str(blockText[col]).lower() in 'abcdefg^_=')):
                col -= 1
        dbg_print ('editor.highlight: row=%d, col=%d' %(blockNumber, col))
        self.book.settledAt.emit(blockNumber+1, col)

        if 0:  # under review sine new approach to syntqx highlighting:
            hi_selection = QtGui.QTextEdit.ExtraSelection()

            hi_selection.format.setBackground(self.palette().alternateBase())
            hi_selection.format.setProperty(QtGui.QTextFormat.FullWidthSelection,
                                            True)
            if self.currentLineColor is not None:
                hi_selection.format.setBackground(self.currentLineColor)
            #setFontUnderline(True)
            hi_selection.cursor = tc
            self.setExtraSelections([hi_selection])
            hi_selection.cursor.clearSelection()

    def handleTextChanged(self):
        self.book.counted = self.book.latency
        dbg_print ('handleTextChanged', self.book.counted)

    def handleLull(self, force=False):
        if force or self.document().isModified():
            dbg_print ("autoSave")
            split = os.path.split(self.fileName)
            fileName = 'autosave_'.join(split)
            self.saveFile(
                fileName=temp_dir+ '/autosave_' + os.path.split(self.fileName)[1])
        tc = self.textCursor()
        position = tc.position()
        if position != self.prevCursorPos:
            self.prevCursorPos = position
            self.highlight(tc)
        if self.highlighter:
            self.highlighter.rehighlight()

    def newFile(self, fileName='new.abc'):
        self.clear()
        self.setFileName(fileName)
        self.book.fileLoaded.emit(self, fileName)

    def closeFile(self):
        self.clear()
        self.haveLoadedFile = False

    def cloneAnyFile(self):
        fileName = QtGui.QFileDialog.getOpenFileName(self,
                                                         "Choose a data file",
                                                         '', '*.abc')[0]
        dbg_print ("cloneAnyFile 2", fileName)
        self.loadFile(fileName, newInstance=True)

    def restart(self):
        self.loadFile(self.fileName)
        sys.exit(0)

    def loadFile(self, fileName, newInstance=None, row=1, col=0):
        dbg_print ("Editor.loadFile", fileName, newInstance, row, col)
        if newInstance is None:
            newInstance = False # self.haveLoadedFile
        if newInstance:
            dbg_print("need to create new instance for", fileName)
            sys.argv[1:] = fileName,
            subprocess.Popen(sys.argv)
            return

        self.setFileName(fileName)
        f = QtCore.QFile(fileName)

        if not f.open(QtCore.QFile.ReadOnly | QtCore.QFile.Text):
            return
        self.highlighter = None  # default, half-expecting to be overwritten by per-extension handler
        self.book.fileLoaded.emit(self, fileName)
        self.readAll(f)
        f.close()
        dbg_print ("Loaded %s" % fileName)
        self.moveToRowCol(row, col)  # primarily to gain focus!
        # self.document().setModified(True) # force rewrite of Score
        self.book.fileSaved.emit(fileName)  # ???

    def setFileName(self, fileName=None):
        if fileName is not None:
            self.fileName = fileName
        title = "%s - %s" % (self.headerText, os.path.abspath(self.fileName))
        dbg_print (title)
        # self.book.dock.setWindowTitle(title)
        self.haveLoadedFile = True
        _dirname, _endname = os.path.split(self.fileName)
        if _dirname:
            os.chdir(_dirname)

    def readAll(self, f):
        dbg_print ('readAll', self, f)
        stream = QtCore.QTextStream(f)
        text = stream.readAll()
        self.setPlainText(text)

    def saveFile(self, fileName=None,):
        self.specialSaveFileName = fileName # None if save is requested by user as opposed to temporary for score generation
        if fileName is None:
            fileName = self.fileName
        if fileName is None:
            return
        #f = QtCore.QFile(fileName)
        out = open(fileName, 'w')
        if not out:
            return
        self.writeAll(out)
        out.close()
        dbg_print ("Saved %s " % fileName)
        self.document().setModified(False)
        self.book.fileSaved.emit(fileName)
        return

    def transpose(self):
        semitones, ok = QtGui.QInputDialog.getInteger(self,
                "Transpose (automatic clef change(s))",
                "semitones (+/- for up/down:)", 0, -24, 24, 1)
        if not ok:
            return
        newFileName, ok  = QtGui.QFileDialog.getSaveFileName(self, "write tansposed to file",
                            "transposed.abc",
                            "(*.abc)")
        if not ok:
            return
        transposedText = Share.abcRaft.abc2abc.process(self.fileName,
                                                transpose=semitones)
        with open(newFileName, 'w') as transposed_file:
            transposed_file.write(transposedText)
        self.book.openThemAll((newFileName,))

    def writeAll(self, out):
        text = self.toPlainText()
        # dbg_print('len(text)=', len(text))
        out.write(text)

    def reloadFile(self):
        dbg_print ("ReloadFile", self.fileName)
        self.loadFile(self.fileName)

    def saveFileAs(self, fileName=None, show=True):
        """
        save the current panel contents to a new file.
        """
        if fileName is None:
            files = QtGui.QFileDialog.getSaveFileName(self,
                "Save source to file as", '', '*.abc')
            if not files:
                return
            fileName = files[0]
        if show:
            self.setFileName(fileName)
        self.saveFile()
        self.book.setTabText(self.book.currentIndex(), os.path.split(fileName)[1])

    def resizeEvent(self,e):
        self.lineNumberArea.setFixedHeight(self.height())
        QtGui.QPlainTextEdit.resizeEvent(self,e)

    def eventFilter(self, object, event):
        if object is self.viewport():
            self.lineNumberArea.update()
            return False
        return QtGui.QPlainTextEdit.eventFilter(object, event)

    def keyPressEvent(self, event):
        """Reimplement Qt method"""
        key = event.key()
        # print (type(event))
        meta = event.modifiers() & QtCore.Qt.MetaModifier
        ctrl = event.modifiers() & QtCore.Qt.ControlModifier
        shift = event.modifiers() & QtCore.Qt.ShiftModifier
        plain = not (meta or ctrl or shift)
        if key == QtCore.Qt.Key_Insert and plain:
            self.setOverwriteMode(not self.overwriteMode())
        if key == QtCore.Qt.Key_Tab and plain and self.highlighter:
            return self.autoComplete(event)
        else:
            QtGui.QPlainTextEdit.keyPressEvent(self, event)

    def autoComplete(self, event):
        print ('autoComplete')
        tc = self.textCursor()
        snippet = self.highlighter.getSnippet(tc)
        for i, piece in enumerate(snippet):
            tc.insertText(piece)
            if i==0:
                pos = tc.position()
        tc.setPosition(pos)
        self.setTextCursor(tc) 

    def getSnippet(self, tc):    #------ Drag and drop
        col0 = col = tc.positionInBlock()
        block = tc.block()
        l = block.length()
        print("ABC get snippet", l)
        blockText = block.text()
        while col and ((col >= (l - 1))
                       or not (str(blockText[col - 1]) in ' |!]')):
            tc.deletePreviousChar()
            col -= 1
        key = blockText[col:col0]
        print("autoComplete key %d:%d '%s'" % (col, col0, key))
        return self.snippets.get(key, ("!%s!" % key,))

    def dragEnterEvent(self, event):
        """Reimplement Qt method
        Inform Qt about the types of data that the widget accepts"""
        source = event.mimeData()
        if source.hasUrls():

            if 1: #mimedata2url(source, extlist=EDIT_EXT):
                print ("dragEnterEvent", "hasUrls")
                event.acceptProposedAction()
            else:
                event.ignore()
        elif source.hasText():
            print ("dragEnterEvent", "hasText")
            event.acceptProposedAction()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        event.acceptProposedAction()

    def dropEvent(self, event):
        """Reimplement Qt method
        Unpack dropped data and handle it"""
        source = event.mimeData()
        if source.hasUrls():
            #paths = map(filenameFromUrl, source.urls())
            paths = [url.path() for url in source.urls()]
            print ("dropEvent", "hasUrls", source.urls(), paths)
            self.book.filenamesDropped.emit(paths)
        elif source.hasText():
            print ("dropEvent", "hasText")
            #editor = self.get_current_editor()
            #if editor is not None:
            #    editor.insert_text( source.text() )
        event.acceptProposedAction()

    def mousePressEvent(self, mouseEvent):
        if (mouseEvent.button() in (QtCore.Qt.LeftButton, QtCore.Qt.RightButton)):
            QtGui.QPlainTextEdit.mousePressEvent(self, mouseEvent)
        print (mouseEvent.button() )
        return


    def wheelEvent(self, event):
        modifiers = QtGui.QApplication.keyboardModifiers()
        if modifiers != QtCore.Qt.ControlModifier:
            return QtGui.QPlainTextEdit.wheelEvent(self, event)

        dbg_print ("Editor.wheelEvent, delta = ", event.delta())
        new_sizeF = self.pointSizeF + (event.delta() / 100.0)
        if new_sizeF > 0:
            self.pointSizeF = new_sizeF
            self.font().setPointSizeF(new_sizeF)
        event.accept()


    class LineNumberArea(QtGui.QWidget):

        def __init__(self, editor):
            QtGui.QWidget.__init__(self, editor)
            self.edit = editor
            self.highest_line = 0
            css = '''
            QWidget {
              font-family: monospace;
              font-size: 10;
              color: black;
            }'''
            self.setStyleSheet(css)
 
        def update(self, *args):
            width = QtGui.QFontMetrics(
                self.edit.document().defaultFont()).width(
                    str(self.highest_line)) + 10
            if self.width() != width:
                self.setFixedWidth(width)
                self.edit.setViewportMargins(width,0,0,0)
            QtGui.QWidget.update(self, *args)
 
        def paintEvent(self, event):
            page_bottom = self.edit.viewport().height()
            font_metrics = QtGui.QFontMetrics(
                self.edit.document().defaultFont())
            current_block = self.edit.document().findBlock(
                self.edit.textCursor().position())
 
            painter = QtGui.QPainter(self)
            painter.fillRect(self.rect(), QtCore.Qt.lightGray)
            
            block = self.edit.firstVisibleBlock()
            viewport_offset = self.edit.contentOffset()
            line_count = block.blockNumber()
            painter.setFont(self.edit.document().defaultFont())
            while block.isValid():
                line_count += 1
                # The top left position of the block in the document
                position = self.edit.blockBoundingGeometry(block).topLeft() + viewport_offset
                # Check if the position of the block is out side of the visible area
                if position.y() > page_bottom:
                    break
 
                # We want the line number for the selected line to be bold.
                bold = False
                x = self.width() - font_metrics.width(str(line_count)) - 3
                y = round(position.y()) + font_metrics.ascent()+font_metrics.descent()-1
                if block == current_block:
                    bold = True
                    font = painter.font()
                    font.setBold(True)
                    painter.setFont(font)
                    pen = painter.pen()
                    painter.setPen(QtCore.Qt.red)
                    painter.drawRect(0, y-14, self.width()-2, 20)
                    painter.setPen(pen)
                    
                # Draw the line number right justified at the y position of the
                # line. 3 is a magic padding number. drawText(x, y, text).
                painter.drawText(x, y, str(line_count))
 
                # Remove the bold style if it was set previously.
                if bold:
                    font = painter.font()
                    font.setBold(False)
                    painter.setFont(font)
 
                block = block.next()
 
            self.highest_line = line_count
            painter.end()
 
            QtGui.QWidget.paintEvent(self, event)



#!/usr/bin/python
"""
Copyright 2015 Hippos Technical Systems BV.

@author: larry
"""
from __future__ import print_function
import sys, os, re, subprocess
from ..share import (Share, QtCore, QtGui, Printer, dbg_print)
from .score import Score
from .syntax import AbcHighlighter

from .external import (Abc2midi, Abcm2svg, Abc2abc)

from .midiplayer import MidiPlayer


class AbcRaft(object):

    midiPlayerExe = 'timidity'

    def __init__(self):
        Share.abcRaft = self
        self.midiPlayer = MidiPlayer()
        self.score = Score()
        self.abc2abc = Abc2abc()
        self.abc2midi = Abc2midi()
        self.abcm2svg = Abcm2svg()

        self.printer = Printer()

        Share.raft.setWindowTitle("Musicraft")
        Share.raft.displayBook.addTab(self.score, "Score")
        Share.raft.editBook.fileLoaded.connect(self.checkLoadedFile)

        self.create_actions()
        self.create_menus()

    def checkLoadedFile(self, editor, filename):
        dbg_print('checkLoadedFile', filename)
        if os.path.splitext(filename)[1] in ('.abc', '.ABC'):
            dbg_print("we expect ABC syntax in " + filename)
            editor.highlighter = syntax.AbcHighlighter(editor.document(), editor)

    def start_midi(self):
        if not (self.abc2midi and self.abc2midi.outFileName):
            return
        if self.midiPlayer:
           self.midiPlayer.play(self.abc2midi.outFileName)

    def pause_midi(self):
        if self.midiPlayer:
           self.midiPlayer.pause()

    def create_actions(self):
        self.start_midi_action = Share.raft.myQAction("Start &Midi",shortcut="Ctrl+M",
                triggered=self.start_midi)

        self.pause_midi_action = Share.raft.myQAction("Pause M&idi",shortcut="Ctrl+,",
                triggered=self.pause_midi)

    def create_menus(self):

        self.midi_menu = QtGui.QMenu("&Midi")
        self.midi_menu.addAction(self.start_midi_action)
        self.midi_menu.addAction(self.pause_midi_action)

        Share.raft.menuBar().addMenu(self.midi_menu)
        Share.raft.menuBar().show()



#!/usr/bin/env python
from __future__ import print_function
"""
Copyright 2015 Hippos Technical Systems BV.
(but borrows somme code from the painting/svgviewer example of PyQt v4.x)

@author: larry
"""

import sys, re
import lxml.etree
import numpy as np

from ..share import (Share, dbg_print, QtCore, QtGui, QtWebKit, WithMenu, Printer)

class TextView(QtGui.QPlainTextEdit, WithMenu):

    menuTag = '&Text'

    def menuItems(self):
        return [
#                    ('Set &Font', 'F', self.changeMyFont,),
        ]

    def __init__(self):
        dbg_print ("TextView.__init__")
        QtGui.QPlainTextEdit.__init__(self)
        WithMenu.__init__(self)

    def showOutput(self, text_bytes):
        self.setPlainText(text_bytes)

    def showAtRowAndCol(self, row, col):
        pass  # for now!

    def locateXY(self, x, y):
        pass  # for now!



#!/usr/bin/python
"""
Copyright 2015 Hippos Technical Systems BV.
Module 'external' within package 'abcraft' relates to the various
command processors (abc2midi etc.) which are executed by abccraft, and to
their assocated widgets and methods.
"""
from __future__ import print_function
import os, re
from ..share import (Share, dbg_print)
from ..raft.external import External


class Abcm2ps(External):
    """
class Abcm2ps - in our usage - relates only to the use of abcm2ps to produce
postscript (as opposed to svg) output. This is currently unused (see
'Abcm2svg' below.)
    """
    fmtNameIn  = '%s.abc'
    fmtNameOut = '%s.ps'
    exec_file = 'abcm2ps'

    def cmd(self, inF, outF, **kw):
        return External.cmd(self, '-A', '-O', outF, inF)


class Abcm2svg(External):
    
    fmtNameIn  = '%s.abc'
    fmtNameOut = '%s_page_.svg'
    exec_file = 'abcm2ps'
    reMsg = r'.*\:(\d+)\:(\d+)\:\s\.*'
    #reMsg = r'.*in\s+line\s(\d+)\.(\d+).*'
    rowColOrigin = (0, 0)

    def __init__(self):
        External.__init__(self)
        self.outFile_CRE = re.compile("Output written on\s+(\S.*)\s\(.*")

    def cmd(self, inF, outF, **kw):
        return External.cmd(self, '-k 640 -v -A -O', outF, inF)

    def fixup(self, output, error):
        # just look for the output file names:
        svgList = []
        for either in (output, error):
            for line in either.split('\n'):
                match = self.outFile_CRE.match(line)
                if match:
                    svgList.append(match.group(1))
        dbg_print (svgList)
        if svgList:
            Share.abcRaft.score.useFiles(svgList)
        return output, error

class Abc2midi(External):

    fmtNameIn  = '%s.abc'
    fmtNameOut = '%s.midi'
    exec_file = 'abc2midi'
    showOut = False
    reMsg = r'.*in\s+line-char\s(\d+)\-(\d+).*'
    rowColOrigin = (0, 0)
    
    def cmd(self, inF, outF, **kw):
        return External.cmd(self, inF, '-v', '-EA', '-o', outF)

    def fixup(self, output, error):
        return '', output+error

class Abc2abc(External):
    
    fmtNameIn  = '%s.abc'
    fmtNameOut = '%.0stransposed.abc'  # sneakily don't use basename at all!
    exec_file = 'abc2abc'
    showOut = False
    reMsg = r'(%Warning|%Error).*'

    def cmd(self, inF, outF, transpose=None, **kw):
        pp = [inF, '-OCC', '-r']
        if transpose is not None:
            pp += ['-t', str(transpose)]
        return External.cmd(self, *pp)

    def fixup(self, output, error):
        new_error = ''
        new_output = ''
        for line in output.split('\n'):
            if self.creMsg.match(line):
                new_output = new_output[:-1]
                new_output = new_output[1:]
                new_error += ('line %d: ' % (1+new_output.count('\n')) + line + '\n')
                
            else:
                new_output += (line + '\n')
        self.stdTab.creMsg = re.compile(r'.*line\s+(\d+)\:.*')
        return new_output, new_error



#!/usr/bin/python
"""
Copyright 2015 Hippos Technical Systems BV.
Module 'external' within package 'abcraft' relates to the various
command processors (abc2midi etc.) which are executed by abccraft, and to
their assocated widgets and methods.
"""
from __future__ import print_function
import logging
logger = logging.getLogger()
import sys, os, re, subprocess, tempfile
from ..share import (Share, dbg_print, QtGui)


class StdTab(QtGui.QPlainTextEdit):
    """
This once very bare looking class is gradually being embellished with facilities for
error location helpers etc. in due course. It is the class behind the several
tabs (Abcm2svg etc.) within the subprocess output notebook.
    """
    def __init__(self, commander):
        QtGui.QPlainTextEdit.__init__(self)
        # self.setFont(commander.font)  # maybe unnecessary - see External.write
        dbg_print (self.__class__.__name__+':__init__... commander.reMsg =',
               commander.reMsg)
        self.creMsg = commander.creMsg
        self.rowColOrigin = commander.rowColOrigin
        self.quiet = False
        self.cursorPositionChanged.connect(self.handleCursorMove)

    def handleCursorMove(self):
        # dbg_print (self.__class__.__name__+':handleCursorMove... self.quiet =',
        #        self.quiet)
        if self.quiet or self.creMsg is None:
            return
        match = self.creMsg.match(self.textCursor().block().text())
        # dbg_print (self.__class__.__name__+':handleCursorMove... match =', match)
        if match is None:
            return
        location = [o1+o2 for (o1, o2) in zip(
                        map(lambda s: int(s), match.groups()),
                       self.rowColOrigin)]

        # print ("Autolocating error in ABC", location )
        
        Share.raft.editBook.moveToRowCol(*location)

    def setPlainText(self, text):
        self.quiet = True
        QtGui.QPlainTextEdit.setPlainText(self, text)
        self.quiet = False


class External(object):
    """
'External' is the generic class representing command processors invoked from
within abcraft.
    """
    fmtNameIn  = '%s.in'
    fmtNameOut = '%s.out'
    #exec_dir = os.path.normpath(os.path.split(__file__)[0] + '/../share/' + os.sys.platform + '/bin').replace(
    #    'linux2', 'linux')
    exec_dir = ''
    exec_file = "base_class_stub_of_exec_file"
    outFileName = None
    showOut = True
    reMsg = None  # r'$^'  # default = don't match any lines.
    rowColOrigin = (0, -1)
    stdFont = 'Courier New', 10, False, False
    useItalic = False
    tabName = True  # = use class name
    lastStdTab = None

    def __init__(self):
        self.font = QtGui.QFont(*self.stdFont)
        self.creMsg = (self.reMsg is not None and re.compile(self.reMsg)) or None
        if self.tabName is True:
            self.tabName = self.__class__.__name__
        if self.tabName:
            External.lastStdTab = self.stdTab = StdTab(self)
            self.stdTab.setFont(self.font)
            Share.raft.stdBook.widget.addTab(self.stdTab, self.tabName)
        elif self.tabName is None:
            self.stdTab = External.lastStdTab
        Share.raft.stdBook.widget.setCurrentWidget(self.stdTab)
        Share.raft.editBook.fileSaved.connect(self.process)

    def cmd(self, *pp):
        answer = ' '.join(((self.exec_dir + self.exec_file),) + pp)
        dbg_print ("External.cmd answer = ", answer)
        return answer

    def handle_output(self, output):
        # maybe don't need this function if it never gets overridden?
        return output

    def process(self, inFileName, **kw):
        #dbg_print(inFileName)
        baseName = os.path.splitext(inFileName)[0]
        if inFileName != (self.fmtNameIn % baseName):
            #logger.warning("ignoring file {0} (doesn't conform to '{1}'".format(
            #                            inFileName,             self.fmtNameIn))
            return
        self.outFileName = self.fmtNameOut % baseName
        if self.cmd is None:
            return
        cmd1 = self.cmd(inFileName, self.outFileName, **kw)
        dbg_print (cmd1)
        outputfile, errorfile = [
            tempfile.NamedTemporaryFile(mode='r+', encoding='utf-8', suffix='.'+self.exec_file+'-'+suffix)
                            for suffix in ('out', 'err')]
        process = subprocess.Popen(cmd1, stdout=outputfile, stderr=errorfile, shell=True)
        process.wait()
        outputfile.seek(0), errorfile.seek(0)
        output, error  = outputfile.read(), errorfile.read()
        outputfile.close(), errorfile.close()
        output, error = self.fixup(output, error)
        self.write(out=self.showOut and output or '', err=error, append=False)
        return output

    def fixup(self, output, error):
        return output, error    # hook function for 're-arranging between output and error.. etc.!

    def write(self, out='', err='', append=True):
        if self.stdTab is None:
            dbg_print ("self.stdTab is None!")
            return
        if not append:
            self.stdTab.setPlainText('')
        self.stdTab.setFont(self.font)  # to cope with stdout/stderr case.
        tc = self.stdTab.textCursor()
        cf = tc.charFormat()
        for blurb, useItalic in ((out, False),(err, True)):
            if blurb in ('', '\n'): # unjustifiable kludge, perhaps .. but it has the desired effect!
                continue            # compensate for extra new line provided by appendPlainText.
            cf.setFontItalic(useItalic)
            tc.setCharFormat(cf)
            self.stdTab.setTextCursor(tc)
            self.stdTab.appendPlainText(blurb)


class StdOut(External):
    tabName = 'System'

class StdErr(StdOut):
    tabName = None  # = hitch-hike with previously created sibling.

    def write(self, out='', err='', append=True):
        return StdOut.write(self, out=err, err=out, append=append)


# -*- coding: utf-8 -*-
"""
Copyright 2015 Hippos Technical Systems BV.
Created on Sun Aug 30 18:18:56 2015

@author: larry
"""
import sys, os, subprocess
from ..share import (Signal, dbg_print, QtCore, QtGui, QtSvg, temp_dir)
from .editor import Editor

class EditBook(QtGui.QTabWidget):
    minimumWidth = 480
    minimumHeight = None
    interval = 100
    latency = 3
    editors = []
    filenamesDropped = Signal(list)
    settledAt = Signal(int, int)
    fileSaved = Signal(str)
    fileLoaded = Signal(QtGui.QPlainTextEdit, str)

# hastily rescued from widgetWithMenu mix-in:
#
    menuTag = '&File'
    whereDockable   = QtCore.Qt.AllDockWidgetAreas
    waitCondition = None
    latency = 8
    counted =0
    fileName = None
    minimumWidth = 640
    minimumHeight = None # 800

    headerText = 'Edit'

    def __init__(self, dock=None):
        dbg_print ("EditBook.__init__", dock)
        self.dock = dock
        QtGui.QTabWidget.__init__(self)
        if self.minimumHeight:
            self.setMinimumHeight(self.minimumHeight)
        if self.minimumWidth:
            self.setMinimumWidth(self.minimumWidth)
        #self.setWidth(640)
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.countDown)
        self.timer.start(self.interval)
        self.filenamesDropped.connect(self.openThemAll)
        self.currentChanged.connect(self.activateCurrent)
        #self.fileLoaded.connect(self.fixTabName)

    #def fixTabName(self, fileName):
    #    pass

    def countDown(self, force=None):
        if force:
            self.counted = force
        if self.counted==0:
            return
        self.counted -=1
        # dbg_print('countDown', self.counted)
        if not self.counted:
            self.activeEdit.handleLull()

    def newFile(self):
        self.openThemAll(force=True)

    def openThemAll(self, filenames=(), force=False):
        if not self.editors:
            force=True
        if force and not filenames:
            filenames = os.path.join(os.getcwd(),'new.abc'),
        dbg_print('openThemAll', filenames)
        if not filenames:
            return
        for fn in filenames:
            #ed = RaftEditor(book=self)
            ed = Editor(book=self)
            self.editors.append(ed)
            self.addTab(ed, os.path.split(fn)[1])
            ed.loadFile(fn)
        self.setActiveEdit(ed)

    def setActiveEdit(self, ed):
        self.activeEdit = ed
        self.setCurrentWidget(ed)
        ed.editBecomesActive.emit()

    def loadAnyFile(self):
        fileName = QtGui.QFileDialog.getOpenFileName(self,
                                                         "Choose a data file",
                                                         '.', '*.*')[0]
#                                                         '.', '*.abc')[0]
        dbg_print ("loadAnyFile 2", fileName)
        self.openThemAll((fileName,))


    def activateCurrent(self, ix):
        dbg_print('activateCurrent', ix)
        self.activeEdit = self.editors[ix]
        # quick fix below needs to be improved!
        if not self.activeEdit.fileName:
            return
        self.activeEdit.setFileName() # to force chdir!
        self.activeEdit.handleLull(force=True)
        # self.activeEdit.editBecomesActive.emit()

    # temporary hacks while getting tabbed approach working:

    def transpose(self):
        self.activeEdit.transpose()

    def reloadFile(self):
        self.activeEdit.reloadFile()

    def saveFile(self):
        self.activeEdit.saveFile()

    def saveFileAs(self):
        self.activeEdit.saveFileAs()

    def closeFile(self):
        self.activeEdit.closeFile()
        self.editors.remove(self.activeEdit)
        self.removeTab(self.currentIndex())
        self.openThemAll()

    def restart(self):
        self.activeEdit.restart()

    def moveToRowCol(self, *location):
        self.activeEdit.moveToRowCol(*location)

    def exit_etc(self):
        while True:
            unsaved = [(editor.specialSaveFileName is not None) for editor in self.editors]
            print ('unsaved', unsaved)
            n_unsaved = sum(unsaved)
            if not n_unsaved:
                break
            ret = QtGui.QMessageBox.warning(self, "Application",
                ("There are unsaved changes in %u document(s).\n" % n_unsaved) +
                "Do you want to save your changes?",
                QtGui.QMessageBox.SaveAll | QtGui.QMessageBox.Discard |
                        QtGui.QMessageBox.Cancel)
            if ret == QtGui.QMessageBox.Cancel:
                return False
            if ret == QtGui.QMessageBox.Discard:
                break
            for editor in self.editors:
                if editor.specialSaveFileName is None:
                    continue
                editor.saveFile()
        sys.exit()



#!/usr/bin/python3
"""
Copyright 2016 Hippos Technical Systems BV.

@author: larry
"""
import sys
from .editbook import EditBook
from ..share import (Share, Signal, dbg_print, QtCore, QtGui, QtSvg, WithMenu)
from .external import (StdOut, StdErr)
#import qdarkstyle

class StdBook(QtGui.QTabWidget):
    headerText = 'error/diagnostic output'
    whereDockable   = QtCore.Qt.AllDockWidgetAreas

    def __init__(self, dock=None):
        QtGui.QTabWidget.__init__(self)

class DisplayBook(QtGui.QTabWidget):
    pass  # for now
class Dock(QtGui.QDockWidget):
    def __init__(self, widgetClass, visible=True):
        QtGui.QDockWidget.__init__(self, widgetClass.headerText)
        self.setAllowedAreas(widgetClass.whereDockable)
        self.widget = widgetClass(dock=self)
        self.setWidget(self.widget)
        self.setVisible(visible)


class Raft(QtGui.QMainWindow, WithMenu):
    menuTag = '&File'

    def __init__(self):
        QtGui.QMainWindow.__init__(self)
        Share.raft = self
        self.resize(1280, 1024)
        self.stdBook = Dock(StdBook,  True)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.stdBook)
        self.stdBook.setMinimumHeight(140)
        self.editBookDock = Dock(EditBook, True)
        # self.raftEditor.setMinimumWidth(320)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.editBookDock)
        self.editBook = self.editBookDock.widget
        sys.stdout = StdOut()
        sys.stderr = StdErr()
        print ("testing stdout...", file=sys.stdout)
        print ("testing stderr...", file=sys.stderr)
        #sys.exit(0)
        self.createMenus()
        self.displayBook = DisplayBook()
        self.setCentralWidget(self.displayBook)
        #self.displayBook.setBaseSize(640, 480)
        #self.displayBook.addTab(QtGui.QGraphicsView(), "test widget")
        WithMenu.__init__(self)


    def start(self):
        self.show()
        self.editBook.openThemAll(sys.argv[1:])

    def about(self):
        QtGui.QMessageBox.about(self, "About 'Raft'",
                "<p>To be updated!.</p>"
                "<p></p>")

    def createMenus(self):
        self.helpMenu = QtGui.QMenu("&Help", self)
        self.aboutAct = self.myQAction("About &Raft", triggered=self.about)
        self.aboutQtAct = self.myQAction("About &Qt", triggered=QtGui.qApp.aboutQt)
        self.helpMenu.addAction(self.aboutAct)
        self.helpMenu.addAction(self.aboutQtAct)
        self.helpAction = self.menuBar().addMenu(self.helpMenu)

    def closeEvent(self, e):
        self.editBook.exit_etc()

    def menuItems(self):
        return [
                    ('&New',           'Ctrl+N', self.editBook.newFile,),
                    ('&Open',          'Ctrl+O', self.editBook.loadAnyFile,),
                    ('&Close',         'Ctrl+C', self.editBook.closeFile,),
                    #('Open in new &Instance', 'Ctrl+I', self.editor.cloneAnyFile,),
                    ('&Reload',        'Ctrl+R', self.editBook.reloadFile,),
                    ('R&estart',       'Ctrl+E', self.editBook.restart,),
                    ('&Save',          'Ctrl+S', self.editBook.saveFile,),
                    ('Save &As',       'Ctrl+A', self.editBook.saveFileAs,),
                    ('E&xit',          'Ctrl+Q', self.editBook.exit_etc),
                    ('&Transpose',     'Ctrl+T', self.editBook.transpose,),
#                    ('Set &Font', 'F', self.changeMyFont,),

        ]

def main(Plugins=()):
    app = QtGui.QApplication(sys.argv)
    # app.setStyleSheet(qdarkstyle.load_stylesheet())
    Raft()
    Share.plugins = []
    for Plugin in Plugins:
        try:
            Share.plugins.append(Plugin())
        except TypeError as exc:
            print(exc, file=sys.stderr)

    Share.raft.start()
    try:
        sys.exit(app.exec_())
    except:
        pass

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Copyright 2015 Hippos Technical Systems BV.
(but borrows some code from the painting/svgviewer example of PyQt v4.x)

@author: larry
"""

import sys, re
import lxml.etree
import numpy as np
from collections import OrderedDict
from ..share import (Share, dbg_print, QtCore, QtGui, QtWebKit, WithMenu)


class MyScene(QtGui.QGraphicsScene):
        
    def mousePressEvent(self, event):
        scP = event.scenePos()
        x = scP.x()
        y = scP.y()
        dbg_print ("MyScene.mousePressEvent: "+
               #event.pos(), event.scenePos(), event.screenPos()
               'scenePos x,y =' + str(x) + ',' + str(y), '  button =' + str(event.button()),
               'scene width =' + str(self.width()) + ' scene height ='+ str(self.height()),
                self.parent())
        if event.button() == 1:
            self.parent().locateXY(x, y)
            event.accept()
        else:
            event.ignore()

    def wheelEvent1(self, event):
        modifiers = QtGui.QApplication.keyboardModifiers()
        if not (modifiers & QtCore.Qt.ControlModifier):
            return  QtGui.QGraphicsScene.wheelEvent(self, event)
        factor = 1.2**( event.delta() / 120.0)
        # self.scale(factor, factor)
        # self.mustApplyTransform = self.transform()
        dbg_print ("MyScene.wheelEvent, delta = ", event.delta())
        event.accept()


class SvgDigest:
    ringColour = 'green'

    locatableTypes = ('N', 'R', 'Z')
    scene = None

    def __init__(self, filename):
        self.svg_file = QtCore.QFile(filename)
        self.quickDic = {}
        self.svg_tree = self.cursorsDad = None
        self.row_col_dict = OrderedDict()
        self.buildQuickDic()

    def AdjustForScene(self, scene):
        self.scene = scene

    def buildQuickDic(self):
        """ extract the all-imortant information from the .svg
            file which enables us to correlate locations within the
            image with locations within the source abc file.
            N.B. row/col/cursor related stuff is being gradually phased
            out from this function.
        """
        fileName = str(self.svg_file.fileName())

        for attr, dtype in (('row',   np.int32), ('col',   np.int32),
                            ('x',     np.float), ('y',     np.float),
                            ('scale', np.float )):
            self.quickDic[attr] = np.array([], dtype=dtype)

        self.svg_tree = lxml.etree.parse(fileName)
        self.abcEltAtCursor = self.eltCursor = dad = None
        eltHead = eltAbc = None
        scale_ = 1.0
        for elt in self.svg_tree.iter():
            if callable(elt.tag):
                continue
            tag_ = elt.tag.split('}')[1]  # get rid of pesky namespace prefix
            if (tag_=='abc'
            and (elt.get('type') in self.locatableTypes)):
                eltAbc = elt # ready to be paired up with a notehead element
            elif tag_=='use':
                attr, val = elt.items()[-1]
                # look for normal note heads and also the special percussion note heads
                if attr.endswith('href'): #  and val.lower() in ('#hd', '#dsh0', '#pshhd', '#pfthd', '#pdshhd', '#pdfthd'):
                    eltHead = elt # ready to be paired up with an 'abc' element
            elif tag_ == 'g':
                tf_ = elt.get('transform')
                if not tf_:
                    continue
                scale_match = re.match('scale\((.*)\)', tf_)
                if scale_match:
                    try:
                        scale_ = float(scale_match.group(1))
                    except ValueError:
                        continue
                    dbg_print("SvgDigest: scale according to g encountered en passant =", scale_)
                continue
            else:
                continue
            if eltAbc is None or eltHead is None:
                continue
# we've 'paired' a note-head and an ABC note description; hurrah!
            sx_ = eltHead.get('x')
            sy_ = eltHead.get('y')
            row_ = int(eltAbc.get('row'))
            col_ = int(eltAbc.get('col'))
            type_ = eltAbc.get('type')
            if not (sx_ and sy_):
                continue
            self.quickDic['x'] = np.append(self.quickDic['x'], scale_ * float(sx_))
            self.quickDic['y'] = np.append(self.quickDic['y'], scale_ * float(sy_))
            self.quickDic['scale'] = np.append(self.quickDic['scale'], scale_)

            self.quickDic['row'] = np.append(self.quickDic['row'], row_)
            self.quickDic['col'] = np.append(self.quickDic['col'], col_)

            self.row_col_dict.setdefault(row_, OrderedDict())[col_] = (eltAbc, eltHead)

            # avoid pairing the same notehead or abc note descripton again!
            eltAbc = eltHead = None
        return

    def insertCursor(self, eltHead, colour='green'):
        self.cursorsDad = eltHead.getparent()
        self.eltCursor = lxml.etree.Element('circle', r='7', stroke=colour,
                          fill="none")
        self.eltCursor.set('stroke-width', '2')
        self.eltCursor.set('cx', eltHead.get('x'))
        self.eltCursor.set('cy', eltHead.get('y'))
        self.cursorsDad.insert(0, self.eltCursor)

        fileName = str(self.svg_file.fileName())
        outFile = open(fileName, 'wb')
        dbg_print('written', fileName)
        self.svg_tree.write(outFile)

    def removeCursor(self):
        if self.eltCursor is not None and self.cursorsDad is not None:
            self.cursorsDad.remove(self.eltCursor)
            self.eltCursor = None

    def rowColAtXY(self, x, y):
        x_dist = x - self.quickDic['x']
        y_dist = y - self.quickDic['y']
        a_dist = x_dist*x_dist + y_dist*y_dist
        am = np.argmin(a_dist)
        row = self.quickDic['row'][am]
        col = self.quickDic['col'][am]
        return row, col
        
class Score(QtGui.QGraphicsView, WithMenu):

    ringColour = 'red'

    menuTag = '&Score'


    def menuItems(self):
        return [
                    ("&Reset Zoom",    'Ctrl+R',      self.resetZoom,),
                    ("&First Page",    'Ctrl+1',      self.showWhichPage,),
                    ("&Next Page",     'Ctrl+PgDown', self.showNextPage,),
                    ("Pre&vious Page", 'Ctrl+PgUp',   self.showPreviousPage,),
                    ('&Print',         'Ctrl+P',      self.printAll,),
                    ('E&xport to PDF', 'Ctrl+Alt+X',  self.printAllToPDF,),
        ]

    def __init__(self):
        dbg_print ("Score.__init__")
        QtGui.QGraphicsView.__init__(self)
        WithMenu.__init__(self)
        self.setTransformationAnchor(QtGui.QGraphicsView.AnchorUnderMouse)
        self.setDragMode(QtGui.QGraphicsView.ScrollHandDrag)
        self.which = 0  # default to show first generated svg until we know better.
        self.fx = self.fy = 0.0
        self.svgDigests = []
        Share.raft.editBook.settledAt.connect(self.showAtRowAndCol)
        Share.abcRaft.midiPlayer.lineAndCol.connect(self.showAtRowAndCol)

        self.svgView = QtWebKit.QGraphicsWebView()
        self.svgView.setFlags(QtGui.QGraphicsItem.ItemClipsToShape)
        self.svgView.setCacheMode(QtGui.QGraphicsItem.NoCache)
        self.svgView.setZValue(0)
        self.prainter = None
        self.svgView.loadFinished.connect(self.svgLoaded)
        self.setScene(MyScene(self))
        self.scene().clear()
        self.scene().addItem(self.svgView)
        self.scene().update()
        dbg_print ("!Score.__init__")

    def svgLoaded(self):
        frame = self.svgView.page().mainFrame()
        fsize = frame.contentsSize()
        self.svgView.resize(QtCore.QSizeF(fsize))
        #print('ensureVisible2 %d %d'  %(self.fx, self.fy))
        #self.ensureVisible(self.fx, self.fy, 1.0, 1.0)
        #self.update()
        if not self.prainter:
            return
        dbg_print ("'prainting' page", self.which)
        if self.which:
            self.printer.newPage()
        else:
            self.prainter.begin(self)
        self.scene().render(self.prainter)
        self.showNextPage()
        if  not self.which:
            self.prainter.end()
            del self.prainter
            self.prainter = None

    def drawBackground42(self, p, rect):
        p.save()
        p.resetTransform()
        p.drawTiledPixmap(self.viewport().rect(),
                self.backgroundBrush().texture())
        p.restore()

    def useFiles(self, filenames=()):
        self.svgDigests = [SvgDigest(filename) for filename in filenames]
        for path in filenames:
            svg_file = QtCore.QFile(path)
            if not svg_file.exists():
                QtGui.QMessageBox.critical(self, "Open SVG File",
                        "Could not open file '%s'." % path)
        self.showWhichPage(self.which, force=True)

    def getEltsOnRow(self, row, which=None):
        if which is True:
            which = self.which
        answer = {}
        for wh, dig in enumerate(self.svgDigests):
            if (which is not None) and (wh!= which):
                continue
            answer.update (self.svgDigests[wh].row_col_dict.setdefault(row, {}))
        return answer

    def showAtRowAndCol(self, row, col):

        Share.eltAbcCursor = None
        dbg_print ('showAtRowAndCol %d %d' %(row, col))
        l = len(self.svgDigests)
        for i in range(l):
            j = (i +self.which) % l
            dictOfRow = self.getEltsOnRow(row, which=j)
            if dictOfRow:
                break
        else:
            dbg_print("can't find page containing svg graphics for row %d"
                      % row)
            return
        self.svgDigests[j].removeCursor()
        for col_ in list(range(col, 0, -1)) + list(range(col, col+5, 1)):
            eltAbc, eltHead = dictOfRow.get(col_, (None, None))
            if isinstance(eltHead, lxml.etree._Element):
                self.fx, self.fy = [float(eltAbc.get(a)) for a in ('x', 'y')]
                self.svgDigests[j].insertCursor(eltHead, colour=self.ringColour)
                break
        else:
            dbg_print("can't find svg graphics correspond to row %d: col %d (page %d)"
                      % (row, col, j))
            self.fx, self.fy = 0.0, 0.0

        # experimental and ugly!
        dbg_print('ensureVisible1 %d %d'  %(self.fx, self.fy))
        self.ensureVisible(self.fx, self.fy, 20.0, 20.0)
        #self.ensureVisible(point.x(), point.y(), 1.0, 1.0)
        self.showWhichPage(j, force=True)
        Share.eltAbcCursor = eltAbc

    def locateXY(self, x, y):
        row, col = self.svgDigests[self.which].rowColAtXY(x, y)
        dbg_print ("locateXY( %d,%d > row,col %d %d" %(x, y, row, col))
        Share.raft.editBook.moveToRowCol(row, col)

    def showNextPage(self):
        dbg_print ('showNextPage')
        return self.showWhichPage(self.which + 1)

    def showPreviousPage(self):
        dbg_print ('showPreviousPage')
        return self.showWhichPage(self.which - 1)

    def showWhichPage(self, which=0, force=False):
        dbg_print ('----- showWhichPage', which)
        which %= len(self.svgDigests)
        if (not force) and (which==self.which):
            return
        svg_file = self.svgDigests[which].svg_file
        if not svg_file.exists():
            raise IOError("'%s' does not exist!" % svg_file.filename())
        self.which = which
        self.svgView.load(QtCore.QUrl(svg_file.fileName()))
        #self.svgDigests[which].AdjustForScene(self.scene())
        self.update()

    def resetZoom(self):
        self.resetTransform()

    def renderAll(self, painter):
        self.prainter = painter
        self.showWhichPage(0, force=True)

    def wheelEvent(self, event):
        dbg_print ("Score.wheelEvent, delta = ", event.delta())
        modifiers = QtGui.QApplication.keyboardModifiers()
        if not (modifiers & QtCore.Qt.ControlModifier):
            # return  self.scrollContentsBy(0, event.delta()*20)
            sbar = self.verticalScrollBar()
            value = sbar.value() - event.delta()
            if event.delta()>0:
                value = max(sbar.minimum(), value)
            else:
                value = min(sbar.maximum(), value)
            sbar.setValue(value)
            dbg_print("%d <= vsbar=%d <= %d" %(sbar.minimum(), value, sbar.maximum()))
            self.update()
            event.accept()
            return
        factor = 1.2**( event.delta() / 120.0)
        self.scale(factor, factor)
        # self.mustApplyTransform = self.transform()
        event.accept()


if __name__ == '__main__':

    class MainWindow(QtGui.QMainWindow):
        """
        warning: not used this '__main__' in months: probably not working!
        """
        def __init__(self):
            super(MainWindow, self).__init__()
    
            self.currentPath = ''
    
            self.score = Score()
    
            fileMenu = QtGui.QMenu("&File", self)
            quitAction = fileMenu.addAction("E&xit")
            quitAction.setShortcut("Ctrl+Q")
    
            self.menuBar().addMenu(fileMenu)
    
    
            quitAction.triggered.connect(QtGui.qApp.quit)
    
            self.setCentralWidget(self.score)
            self.setWindowTitle("SVG Viewer")
            self.score.useFiles(sys.argv[1:] or
                ['test.svg'])
            self.score.resize(self.score.sizeHint() + QtCore.QSize(
                80, 80 + self.menuBar().height()))


    app = QtGui.QApplication(sys.argv)
    mainWindow = MainWindow()
    mainWindow.show()
    sys.exit(app.exec_())



#!/usr/bin/python
"""
Copyright 2015 Hippos Technical Systems BV.

@author: larry
"""
from __future__ import print_function
import sys, os, re, subprocess
from ..share import (Share, dbg_print, QtCore, QtGui, Printer)
from .html_view import HtmlView
from .text_view import TextView
from .external import Python
from .syntax import PythonHighlighter


class PyRaft(object):

    def __init__(self):
        Share.pyRaft = self
        self.htmlView = HtmlView()
        self.textView = TextView()
        self.python = Python()
        Share.raft.displayBook.addTab(self.htmlView, "Html")
        Share.raft.displayBook.addTab(self.textView, "Text")
        Share.raft.editBook.fileLoaded.connect(self.checkLoadedFile)

    def checkLoadedFile(self, editor, filename):
        dbg_print('checkLoadedFile', filename)
        if os.path.splitext(filename)[1] in ('.py', '.pyw'):
            dbg_print(filename + "  ... is one of mine!")
            editor.highlighter = syntax.PythonHighlighter(editor.document())



#!/usr/bin/python
"""
# terpsichore - another object oriented approach to music representation in the python language.
#    The defining feature of 'terpsichore' iis that notes are represented not as quoted strings
#    but as instances of a 'note' class.
# N.B. This has been hacked rather impatiently/unsympathetically in order to use it within
#    'musicraft'.
"""
import numpy, math
# import analyse

all_Keys = {}
note_letters = ('C', None, 'D', None, 'E', 'F', None, 'G', None, 'A', None, 'B')

notes_by_Pitch = numpy.array([[None,None]]*128)

class musicGroup(object):
    pass

class musicItem(object):
    pass

class Note_(musicItem):

    def __init__(self, frequency=None, pitch=None, octave=None, length=None, instrument=None, real_name=None):
        if frequency is not None:
            aPitch = 69 + 12 * math.log((frequency / 440.0), 2.0)
            octave = int(aPitch/12)-1
            pitch = aPitch - 12*(octave+1)
            print (aPitch, pitch, octave)
        self.pitch = pitch
        self.octave = octave
        self.length = length
        self.real_name = real_name

    def __repr__(self):
        return "Note:'%s' (pitch=%g)"%(self.real_name, self.GetPitch())

    def GetPitch(self, octave=None):
        
        octave = octave or self.octave
        if octave is None:
            return None
        return (octave+1)*12+self.pitch
    
class Note(musicGroup):
    pass

AllNotes = {}

class _Mode(musicItem):
    def __init__(self, isMinor, suffix, full_name):
        self.isMinor = isMinor
        self.suffix = suffix
        self.full_name = full_name

    def __repr__(self):
        return "Mode:'%s' (suffix=%s)"%(self.full_name, self.suffix)

class Mode:
    Major = _Mode(False, 'M', 'Major')
    Minor = _Mode(True, 'm', 'Minor')

class _Key(object):
    # not fully thought out yet!
    def __init__(self, fifths, mode, real_name):
        self.fifths = fifths
        self.real_name = real_name
        self.mode = mode

    def __repr__(self):
        return "Key:'%s' (fifths=%d)\n" %(self.real_name, self.fifths)

class Key(musicGroup):
    pass

for octave in [None]+list(range(-1,10)):
    #print ("------- octave........", octave)
    for pitch, letter in enumerate(note_letters):
        if not letter:
            continue
        if octave is None:
            octave_suffix = ''
        else:
            octave_suffix = '%d' %octave
        for delta, accidental in ((-2, 'bb'), (-1, 'b'), (0, ''), (1, '#'), (2, 'x')):
            real_name = letter+accidental+octave_suffix
            python_name = real_name.replace('#', 'i')
            stepped_pitch = pitch+delta
            new_note = Note_(pitch=stepped_pitch, octave=octave, real_name=real_name)
            #print ('new_note', "=", new_note)
            new_Pitch = new_note.GetPitch()
            if (new_Pitch is not None) and (new_Pitch<0 or new_Pitch>=128):
                continue
            setattr(Note, python_name, new_note)
            AllNotes[new_Pitch] = new_note
            #print ("new_Pitch =", new_Pitch)
            if delta in (2, -2):
                continue
            for i_sharp_flat in (0,1):
                if octave is None:
                    # convenient to hang the key creation stuff in here:
                    for mode in (Mode.Major, Mode.Minor):
                        python_key_name = python_name+mode.suffix
                        if python_key_name in all_Keys:
                            continue
                        real_key_name = real_name+mode.suffix
                        if mode.isMinor:
                            fifths = stepped_pitch+3
                        else:
                            fifths = stepped_pitch
                        if fifths & 1:
                            fifths = fifths+6
                        fifths = ((fifths+6)%12)-6
                        if (not i_sharp_flat) and (fifths<0):
                            fifths += 12
                        elif i_sharp_flat and (fifths>0):
                            fifths -= 12
                        setattr(Key, python_key_name, _Key(fifths, mode, real_key_name))
                elif (2*i_sharp_flat != (1+delta) and
                          (notes_by_Pitch[new_Pitch][i_sharp_flat] is None
                        or len(notes_by_Pitch[new_Pitch][i_sharp_flat].real_name) > 2)):
                    notes_by_Pitch[new_Pitch][i_sharp_flat] = new_note

class _Clef(musicItem):
    def __init__(self, lines=[], marked=None, symbol=None, scaleHint=16, descent=7):
        self.lines = lines
        self.marked = marked
        self.symbol = symbol
        self.scaleHint = scaleHint
        self.descent = descent

class Clef(musicGroup):
    Treble = _Clef(lines=(Note.E4, Note.G4, Note.B4, Note.D5, Note.F5),
                   descent=5, marked=1, symbol='Treble_clef_inv.png', scaleHint=16)
    Bass   = _Clef(lines=(Note.G2, Note.B2, Note.D3, Note.F3, Note.A3),
                   descent=8.5, marked=3, symbol='Bass_clef_inv.png', scaleHint=13)

class Voice(musicItem):
    def __init__(self, pitch_shift=0, clef=None, key=Key.CM, default_octave=3,
                    length=None, instrument=None, name=None):
        self.pitch_shift = pitch_shift
        self.clef = clef
        self.key = key
        self.default_octave = default_octave
        self.instrument = instrument
        self.name = name

    def __repr__(self):
        return "voice:'%s' (pitch_shift=%u)" %(self.name, self.pitch_shift)

    def GetPitch(self, wild, transpose=False):
        if isinstance(wild, (tuple, list, numpy.ndarray)):
            #print("list or tuple!")
            return [self.GetPitch(item, transpose) for item in wild]
        elif isinstance(wild, Note_):
            #print("note!")
            wild = wild.GetPitch()
            if transpose:
                wild -= self.pitch_shift
        return wild  # number, presumably!

    def GetNote(self, note_or_number, transpose=False):
        if isinstance(note_or_number, Note_):
            return note_or_number # note, thus
        elif isinstance(note_or_number, int):
            return notes_by_Pitch[note_or_number][self.key.fifths <= 0]
        elif isinstance(note_or_number, float):
            #return notes_by_Pitch[int(note_or_number)][self.key.fifths<0]
            i_near = int(note_or_number+0.5)
            #return str(i_near)
            note_near = notes_by_Pitch[i_near][self.key.fifths<0]
            miss = note_or_number - i_near
            if miss > 0.1:
                s_miss = '+'
            elif miss < -0.1:
                s_miss = '-'
            else:
                s_miss = '='
            return Note_(octave=note_near.octave, pitch=note_near.pitch+miss,
                        real_name=note_near.real_name+s_miss)

    #def DeriveNote(self, samples, sampleRate=44100, minFreq=10, maxFreq=2000, spread=3,         
    def DeriveNote(self, samples, sampleRate=44100, minFreq=10, maxFreq=400, spread=3,         
                            transpose=False, concertPitch=440.0):
            nSamples = len(samples)
            fftAmplitudes = (abs(numpy.fft.fft(samples))/nSamples)[minFreq:maxFreq]
            fftFrequencies = (sampleRate*numpy.arange(nSamples))[minFreq:maxFreq]/(nSamples*1)
            indexMaxAmplitude = numpy.argmax(fftAmplitudes)
            fftAmplitudes = fftAmplitudes[indexMaxAmplitude-spread:indexMaxAmplitude+spread+1]
            fftFrequencies = fftFrequencies[indexMaxAmplitude-spread:indexMaxAmplitude+spread+1]
            total = numpy.sum(fftAmplitudes)
            #print ("total", total)
            volume = math.log(total+1, 2.0)
            avgFrequency = numpy.sum(fftFrequencies*fftAmplitudes)/total
            absPitch = 69 + 12 * math.log((avgFrequency / concertPitch), 2.0)
            # print  ("absPitch =", absPitch)
            try:
                avgNote = self.GetNote(absPitch, transpose=transpose)
            except:
                avgNote = None
            return avgNote, volume
        
current_voice = default_voice = Voice()

class _Instrument(musicItem):
    """ distinction between instrument and voice rather iffy at present!
    """
    def __init__(self, clef=Clef.Treble, strings=None, real_name=None):
        self.strings = strings
        self.real_name = real_name
        self.clef = clef

    def __repr__(self):
        s = self.strings and "(%s strings)" %len(self.strings) or ""
        return "instrument:'%s'%s" %(self.real_name,s)
class Instrument(musicGroup):
    pass

def add_Instrument(real_name='ANInstrumment', python_name=None, clef=Clef.Treble, strings=None):
    if not python_name:
        python_name = real_name.replace(' ', '_').replace('#', 'i')
    setattr(Instrument, python_name, _Instrument(real_name=real_name, clef=clef, strings=strings))
    
def isInstrument(entity):
    return isinstance(entity, _Instrument)

add_Instrument('Guitar', strings=(Note.E2, Note.A2, Note.D3, Note.G3, Note.B3, Note.E4))
add_Instrument('Violin', strings=(Note.G3, Note.D4, Note.A4, Note.E5))
add_Instrument('Viola',  strings=(Note.C3, Note.G3, Note.D4, Note.A4))
add_Instrument('Chello', strings=(Note.C2, Note.G2, Note.D3, Note.A3))

# following two 'instruments' are dubious but handy at this stage of development:

add_Instrument('Treble Voice')
add_Instrument('Bass Voice', clef=Clef.Bass)

current_instrument = Instrument.Guitar

# quick hack for use by fortuna440

def best_note_name_for_pitch(i):
    nbp = notes_by_Pitch[i]
    return min ([n.real_name for n in nbp if not '#' in n.real_name])
#print(Instrument.__dict__.values())

if __name__ == '__main__':
    #print (Note.A3, Note.B4, [(notes_by_Pitch[i],'\n') for i in range(56,62)])
    print ([best_note_name_for_pitch(i) for i in range(50,70)])
    clef =Clef.Bass
    marked_semi = clef.lines[clef.marked]
    print(marked_semi, marked_semi.GetPitch() - clef.descent)  #, dir(marked_semi))



#!/usr/bin/python2.7
from __future__ import print_function
import sys, os, tempfile

dbg_print = (int(os.getenv('ABCRAFT_DBG', 0)) and print) or (lambda *pp, **kw: None)

class Share:
    eltAbcCursor = None

image_dir = os.path.normpath(os.path.split(__file__)[0] + '/..') + '/images/'
temp_dir = tempfile.gettempdir()

qt_module_name = os.getenv('ABCRAFT_QT', 'PySide')
if qt_module_name == 'PySide':
    from PySide import (QtCore, QtGui, QtSvg, QtWebKit)
    Signal = QtCore.Signal
    dbg_print ("using PySide!")
    name_from_dialog = lambda x:x[0]
elif qt_module_name == 'PyQt4':
    from PyQt4 import (QtCore, QtGui, QtSvg, QtWebKit)

    Signal = QtCore.pyqtSignal
    dbg_print("using Pyqt4!")
    name_from_dialog = lambda x:x
elif qt_module_name == 'PyQt5':
    from PyQt5 import (QtCore, QtGui, QtSvg, QtWebKit)

    Signal = QtCore.pyqtSignal
    dbg_print("using Pyqt4!")
    name_from_dialog = lambda x:x
else:
    raise NameError("bad value: ABCRAFT_QT = " + qt_module_name)


class Printer(QtGui.QPrinter):
    pageSize = QtGui.QPrinter.A4

    def __init__(self):
        dbg_print ("Printer.__init__")
        QtGui.QPrinter.__init__(self, QtGui.QPrinter.HighResolution)
        self.setPageSize(self.pageSize)
        dbg_print ("!Printer.__init__")


class WithMenu(object):
    menuTag = None

    def __init__(self):
        self.printer = Printer()
        self.compositeName = 'temp'  # under review, like so much!
        self.menu = QtGui.QMenu(self.menuTag)
        if not (self.menuTag and self.menuItems()):
            return
        for tag, shortcut, func in self.menuItems():
            action = self.myQAction(tag, shortcut=shortcut, triggered=func)
            self.menu.addAction(action)
        Share.raft.menuBar().addMenu(self.menu)

    def menuItems(self):
        return [
        ]

    def myQAction(self, menuText, shortcut=None, triggered=None, enabled=None,
                  checkable=None, checked=None):
        action = QtGui.QAction(menuText, self)
        if shortcut:
            action.setShortcut(shortcut)
        if triggered:
            action.triggered.connect(triggered)
        if enabled is not None:
            action.setEnabled(enabled)
        if checkable is not None:
            action.setCheckable(checkable)
        if checked is not None:
            action.setChecked(checked)
        return action

    def printAll(self, toPDF=False):
        # fileName = self.compositeName +'.pdf'
        fileName = os.path.splitext(Share.raft.editBook.activeEdit.fileName)[0] + '.pdf'
        print(fileName)
        if toPDF:
            files = QtGui.QFileDialog.getSaveFileName(self,
                "write PDF to file", fileName, '*.pdf')
            fileName = files[0]
            if not fileName:
                return
        self.printer.setDocName(fileName)
        self.printer.setOutputFileName(toPDF and fileName or '')
        self.renderAll(QtGui.QPainter(self.printer))

    def renderAll(self, painter):
        self.scene().render(painter)

    def printAllToPDF(self):
        self.printAll(toPDF=True)




#!/usr/bin/python
"""
Copyright 2015 Hippos Technical Systems BV.

@author: larry
"""
from __future__ import print_function
import sys
import numpy
import pyaudio
from ..share import (Share, dbg_print, QtCore, QtGui, Printer)
from ..raft.external import StdTab
from .terpsichore import default_voice
from .pitch_view import PitchView


class FreqRaft(object):
    inputDeviceIndex = 0
    creMsg = reMsg = None  # to be reviewed - nec. for StdTab
    stdFont = 'Courier New', 10, False  # likewise!
    rowColOrigin = (0, -1)  # also likewise!

    def __init__(self):
        Share.freqRaft = self
# following code cribbed from source code of myPlot1.py:
        self.pitch_view = PitchView()
        Share.raft.displayBook.addTab(self.pitch_view, "Tuning")
        self.stdTab = StdTab(self)
        Share.raft.stdBook.widget.addTab(self.stdTab,
                                     'frequency')

    def write(self, text):
        self.stdTab.appendPlainText(str(text))
        #print(text)




"""
This ABC syntax highlighter doesn't need to do much parsing; this is done by
'abcm2ps' and written into the '.svg' file. 'score.py' then constructs an easy to
use dictionary structure from this,
"""
#built using python highligher as starting point.

from ..share import (dbg_print, QtCore, QtGui, Share)

import sys


def format(color, style='', size=None):
    """Return a QtGui.QTextCharFormat with the given attributes.
    """
    _color = QtGui.QColor()
    _color.setNamedColor(color)

    _format = QtGui.QTextCharFormat()
    _format.setForeground(_color)
    if 'bold' in style:
        _format.setFontWeight(QtGui.QFont.Bold)
    if 'italic' in style:
        _format.setFontItalic(True)
    if size is not None:
        _format.setFontPointSize(size)
    return _format


# Syntax styles that can be shared by all languages
STYLES = {
    'default': format('blue'),
    'notename': format('black', 'bold'),
    'string':  format('cyan', 'italic'),
    'comment':  format('green', 'italic'),
    'header': format('red', 'italic'),
    'decorator': format('black', 'italic'),
    'command': format('black', 'italic'),
}


class AbcHighlighter (QtGui.QSyntaxHighlighter):
    """Syntax highlighter for the abc(plus) music description language.
    """
    # Python keywords
    def __init__(self, document, editor):
        QtGui.QSyntaxHighlighter.__init__(self, document)
        self.editor = editor

    def highlightBlock(self, text):
        """Apply syntax highlighting to the given block of text.
        """

        # I seem to have acquiesced to abcm2ps's scheme of counting rows from 1:
        elts_on_cols_in_row = Share.abcRaft.score.getEltsOnRow(1+self.currentBlock().blockNumber())
        for key_ in STYLES.keys():
            STYLES[key_].setFontPointSize(self.editor.pointSizeF)
        span = len(text)
        if text[1:2]==':':
            style_name = "header"
        elif text[0:2] == '%%':
            style_name = "command"
        else:
            ix =0
            while ix < len(text):
                style_name = 'default'
                span = 1
                eltAbc, eltHead = elts_on_cols_in_row.get(ix, (None, None))
                if eltAbc is not None:
                    style_name = 'notename'
                elif text[ix]=='%':
                    style_name = 'comment'
                    span = len(text)-ix
                else:
                    for startChar, endChar, maybe_style_name in (
                        ('"', '"', 'string'),
                        ('[', ']', 'header'),
                        ('!', '!', 'decorator'),
                    ):
                        if text[ix]!=startChar:
                            continue
                        try:
                            jx = text[ix+1:].index(endChar)
                        except ValueError:
                            break
                        #  broken for now! if jx>2 and text[ix+2]==':':  # distinguish embedded commands from chords
                        style_name = maybe_style_name
                        span = jx + 2
                        break
                self.setFormat(ix, span, STYLES[style_name])
                ix += span
        self.setCurrentBlockState(0)

    snippets = {
        'V': ('V:', ' name="', '" sname="', '"\n',),  # new voice
        'Q': ('Q:1/4',),  # new tempo indication
        '12': ('[1 ', ' :| [2 ',),  # varied repeat ending coding
        'cr': ('!<(!', ' !<)!',),  # hairpin dynamic
        'di': ('!>(!', '!>)!',),  # hairpin dynamic
        'CR': ('"_cresc."',),
        'Cr': ('"^cresc."',),
        'MR': ('"_molto rit."',),
        'Mr': ('"^molto rit."',),
        'PR': ('"_poco rit."',),
        'Pr': ('"^poco rit."',),
        'SB': ('"_steady beat"',),
        'Sb': ('"^steady beat"',),
        'm': ('[M:', '2/', '4]',),  # mid-line time-sig change
        'tt': ('!tenuto!',),
        'tp': ('!teepee!',),
        'ac': ('!>!',),  # accent; '><TAB>' also works
        'ro': ('!///!',),  # roll/roffel; '///<TAB>' also works
        'st': ('!dot!',),  # staccato; 'dot<TAB>' also works
        '.': ('!dot!',),  # staccato; 'dot<TAB>' also works
        'gl': ('!-(!', '!-)!'),  # glissando
        'sd': ('[I:pos stem down]',),
        'su': ('[I:pos stem up]',),
        'sa': ('[I:pos stem auto]',),
    }

    def getSnippet(self, tc):    #------ Drag and drop
        col0 = col = tc.positionInBlock()
        block = tc.block()
        l = block.length()
        print("ABC get snippet", l)
        blockText = block.text()
        while col and ((col >= (l - 1))
                       or not (str(blockText[col - 1]) in ' |!]')):
            tc.deletePreviousChar()
            col -= 1
        key = blockText[col:col0]
        print("autoComplete key %d:%d '%s'" % (col, col0, key))
        return self.snippets.get(key, ("!%s!" % key,))




#!/usr/bin/python
# -*- encoding: utf8 -*-
"""
Copyright 2015 Hippos Technical Systems BV.
Module 'external' within package 'abcraft' relates to the various
command processors (abc2midi etc.) which are executed by abccraft, and to
their assocated widgets and methods.
"""
from __future__ import print_function
import os, re
from ..share import (Share, dbg_print)
from ..raft.external import External

HTML_PREAMBLE = "Content-type: text/html"


class Python(External):
    """
class Python -
    """
    fmtNameIn  = '%s.py'
    fmtNameOut = '%s.html'  # ? maybe unused
    exec_dir = '/usr/bin/'
    exec_file = 'python3'

    def cmd(self, inF, outF, **kw):
        print(outF)
        Share.pyRaft.htmlView.fileName = outF  # quick and dirty fix!
        return External.cmd(self, inF)

    def handle_output(self, output):
        if output.startswith(HTML_PREAMBLE):
            Share.pyRaft.htmlView.showOutput(output[output.index('<html'):])
        else:
            Share.pyRaft.textView.showOutput(output)
        return output



"""
MusicRaft
"""

#from distutils.core import setup
from setuptools import setup

setup(name = 'MusicRaft',
    version = '0.5.0',
    author = "Larry Myerscough",
    author_email='hippostech@gmail.com',
    packages=['musicraft', 'musicraft.abcraft', 'musicraft.freqraft', 'musicraft.pyraft', 'musicraft.raft', 'musicraft.share', ],
    scripts=['bin/run_musicraft.py'],
    url='http://larry.myerscough.nl/terpsichore/musicraft.py',
    license='LICENSE.txt',
    description='GUI for abcplus music notation.',
    long_description=open('README.txt').read(),
    install_requires=[
        "numpy >= 1.1.1",
        "mido == 1.1.14",
        "pyqtgraph >= 0.10.0",
# problematic!        "PySide >= 1.1.1",

    ],
)



#!/usr/bin/env python3
"""
Copyright 2015 Hippos Technical Systems BV.

@author: larry
"""
from __future__ import print_function
import sys
import musicraft
print ("imported musicraft package from", musicraft.__file__)

# Below are examples of how to 'doctor' the behaviour of musicraft.
# This can be handy if e.g. you've installed a newer version of abcm2ps than that on the standard path.
#
#musicraft.abcraft.external.Abcm2svg.exec_dir = '/usr/local/bin/'
#musicraft.abcraft.external.Abc2midi.exec_dir = '/usr/local/bin/'
#musicraft.abcraft.external.Abc2midi.reMsg = r'.*in\s+line-char\s(\d+)\-(\d+).*'
#
# uncomment (and maybve adjust) the above lines only if you need to 'doctor' the behaviour of musicraft.

# call the 'raft' with just the 'musicraft' plugin; other optional experimental plugins are currently disabled.
#
musicraft.raft.main(
    Plugins=(musicraft.abcraft.AbcRaft,
             musicraft.pyraft.PyRaft,
           #  musicraft.freqraft.FreqRaft,
             )
)
